#include <sstream>

#include <cstdlib>
#include <cstring>

#include <gtkmm.h>

#include "gui-main.h"
#include "parse.h"

#include <iostream>

#include <readline/history.h>

#include <gtkmm.h>

static Glib::RefPtr<Gtk::TextBuffer::Mark> beg_mark;

static char *
strsave (const char *s)
{
  if (! s)
    return 0;

  int len = strlen (s);
  char *tmp = new char [len+1];
  tmp = strcpy (tmp, s);
  return tmp;
}

class command_window : public Gtk::Window
{
public:

  command_window (void)
    : m_VBox (Gtk::ORIENTATION_VERTICAL), scrolled_window (),
      m_TextView (), buffer ()
  {
    set_title ("Gtk::TextView example");
    set_border_width (5);
    set_default_size (600, 400);

    add (m_VBox);

    // Add the TreeView, inside a ScrolledWindow, with the button
    // underneath:

    scrolled_window.add (m_TextView);

    // Only show the scrollbars when they are necessary:

    scrolled_window.set_policy (Gtk::POLICY_AUTOMATIC, Gtk::POLICY_ALWAYS);

    m_VBox.pack_start (scrolled_window);

    buffer = Gtk::TextBuffer::create ();
    buffer->set_text
      ("Gtk Example Calculator.\n"
       "Available operations: + - * / ^ ()\n"
       "Semicolon terminates statement.\n"
       "Up Arrow key moves to previous line in the command history.\n"
       "Down Arrow key moves to next line in the comand history.\n\n");
    m_TextView.set_buffer (buffer);

    do_prompt ();

    show_all_children ();
  }

  virtual ~command_window (void) { }

protected:

  bool on_key_press_event (GdkEventKey* event)
  {
    bool retval = false;

    if (! event)
      return retval;

    if (event->type == GDK_KEY_PRESS)
      {
        retval = true;

        // FIXME -- of course, this should be done with a dispatch
        // table so that functions can be easily rebound to different
        // keys...

        switch (event->keyval)
          {
          case GDK_KEY_Return:
            accept ();
            break;

          case GDK_KEY_Up:
          case GDK_KEY_Down:
            history (event->keyval == GDK_KEY_Up);
            break;

          default:
            retval = false;
          }

      }

    if (! retval)
      Gtk::Widget::on_key_press_event (event);

    return retval;
  }

  // Accept an input line, parse and possibly execute it.

  void accept (void)
  {
    std::string line = get_current_line ();

    int status = parse_and_execute (line);

    buffer->insert_at_cursor ("\n");

    scroll_to_bottom ();

    insert_result ();

    do_prompt ();
  }


  // Retrieve a command from the history list and insert it on the
  // current command.

  void history (bool up)
  {
    HIST_ENTRY *entry = up ? previous_history () : next_history ();

    if (entry)
      {
        erase_line ();

        buffer->insert_at_cursor (entry->line);
      }
    else if (! up)
      erase_line ();

    scroll_to_bottom ();
  }

  void erase_line (void)
  {
    Glib::RefPtr<Gtk::TextBuffer::Mark> mark = buffer->get_insert ();

    Gtk::TextBuffer::iterator beg_iter = buffer->get_iter_at_mark (beg_mark);

    buffer->erase (beg_iter, buffer->end ());
  }

  std::string get_current_line (void)
  {
    Glib::RefPtr<Gtk::TextBuffer::Mark> mark = buffer->get_insert ();

    Gtk::TextBuffer::iterator beg_iter = buffer->get_iter_at_mark (beg_mark);
    Gtk::TextBuffer::iterator iter = buffer->get_iter_at_mark (mark);

    Glib::ustring tmp = buffer->get_text (beg_iter, iter);

    return tmp;
  }

  void do_prompt (void)
  {
    Glib::RefPtr<Gtk::TextBuffer::Mark> mark = buffer->get_insert ();

    const char *prompt = beg_of_stmt ? ">> " : "";

    Gtk::TextBuffer::iterator iter = buffer->get_iter_at_mark (mark);

    iter = buffer->insert (iter, prompt);

    beg_mark = buffer->create_mark (iter);

    scroll_to_bottom ();
  }

  void insert_result (void)
  {
    if (have_result)
      {
        std::ostringstream buf;
        buf << "ans = " << result << "\n";

        buffer->insert_at_cursor (buf.str ());

        scroll_to_bottom ();

        have_result = false;
      }
  }

  void scroll_to_bottom (void)
  {
    Glib::RefPtr<Gtk::Adjustment> adj = scrolled_window.get_vadjustment ();

    adj->set_value (adj->get_upper ());

    while (Gtk::Main::events_pending ())
      Gtk::Main::iteration ();
  }

  // Child widgets:

  Gtk::VBox m_VBox;

  Gtk::ScrolledWindow scrolled_window;
  Gtk::TextView m_TextView;

  Glib::RefPtr<Gtk::TextBuffer> buffer;
};

int
gui_main (int argc, char *argv[])
{
  parser_init ();

  Gtk::Main kit (argc, argv);

  command_window window;

  // Shows the window and returns when it is closed.
  Gtk::Main::run(window);

  parser_fini ();

  return 0;
}
